// Transforming file increase Contrast and Saturation twice
#define _WIN32_WINNT 0x0501

#pragma warning (disable: 4312)
#pragma warning (disable: 4996)

//#define ALLEGRO_USE_CONSOLE 1

#include <allegro.h>
#include <winalleg.h>
#include <vector>
#include <map>
#include <math.h>
#include "string_conv.h"
#include "CommandLineParser.h"
#include <string>

using namespace std;
using namespace Epoch::Foundation;

#define DOUBLE_MAX 99999999999

FILE *fp;

BITMAP *bmp1;
BITMAP *out;
CommandLineParser parser; 

struct rgb {
	int b;
	int g;
	int r;
	bool operator==(const rgb &ar)
	{
		if (r==ar.r && g==ar.g && b==ar.b)
			return true;
		else
			return false;
	}
};

// simple sort to distinguish different colors for maps, sets etc
bool operator<(const rgb &l, const rgb &r) 
{
	if( l.r == r.r && l.g == r.g && l.b < r.b) return true;
	if( l.r == r.r && l.g < r.g) return true;
	if( l.r < r.r ) return true;
	return false;
}

vector < rgb > atari_palette;
vector < rgb > colors1;
vector < rgb > colors2;

inline void rgb_putpixel(BITMAP *bmp,int x,int y, rgb &col)
{
	putpixel(bmp,x,y,makecol(col.r,col.g,col.b));
}

void error(char *e)
{
	allegro_message(e);
	allegro_exit();
	exit(1);
}

void error(char *e, int i)
{
	allegro_message("%s %d",e,i);
	allegro_exit();
	exit(1);
}

double RGBApproximationDistance(rgb &e1, rgb &e2)
{
	long rmean = ( (long)e1.r + (long)e2.r ) / 2;
	long r = (long)e1.r - (long)e2.r;
	long g = (long)e1.g - (long)e2.g;
	long b = (long)e1.b - (long)e2.b;
	return sqrt((double)(((512+rmean)*r*r)>>8) + 4*g*g + (((767-rmean)*b*b)>>8));
}

double RGBEuclidianDistance(rgb &col1, rgb &col2)
{
	double distance=0;
	// euclidian distance
	distance+=((double)col1.r-(double)col2.r)*((double)col1.r-(double)col2.r);
	distance+=((double)col1.g-(double)col2.g)*((double)col1.g-(double)col2.g);
	distance+=((double)col1.b-(double)col2.b)*((double)col1.b-(double)col2.b);
	return sqrt(distance);
}


bool LoadAtariPalette(const char *filename)
{
	size_t i;
	rgb col;
	FILE *fp=fopen(filename,"rb");
	if (!fp)
	{
		error("Error opening .act palette file!\n");
		exit(0);
	}
	for (i=0;i<256;++i)
	{
		col.r=fgetc(fp);
		col.g=fgetc(fp);
		col.b=fgetc(fp);
		atari_palette.push_back(col);
	}
	fclose(fp);

	for (i=0;i<16;++i)
	{
		colors1.push_back(atari_palette[i]);
		colors2.push_back(atari_palette[i*16+6]);
	}
	colors2[0]=atari_palette[0];
	return true;
}

pair <unsigned char,unsigned char> FindClosestColorIndexesToRepresent(rgb &col1, rgb &col2)
{
	pair <unsigned char,unsigned char> result;

	double distance;
	double min_distance=DOUBLE_MAX;
	rgb col,avg;

	col=col1;
	col.r+=col2.r;
	col.g+=col2.g;
	col.b+=col2.b;
	col.r/=2;
	col.g/=2;
	col.b/=2;

	int a,b;
	for(a=0;a<16;++a)
	{
		for(b=0;b<16;++b)
		{
			avg=colors1[a];
			avg.r+=colors2[b].r;
			avg.g+=colors2[b].g;
			avg.b+=colors2[b].b;
			avg.r/=2;
			avg.g/=2;
			avg.b/=2;
			distance=RGBApproximationDistance(col,avg);
			if (distance<min_distance)
			{
				min_distance=distance;
				result.first=a;
				result.second=b;
			}
		}
	}
	return result;
}

inline unsigned char FindAtariColorIndex(rgb &col)
{
	size_t i;
	// Find the most similar color in the Atari Palette
	unsigned char most_similar=0;
	double distance;
	double min_distance=DOUBLE_MAX;
	for(i=0;i<256;++i)
	{
		distance=RGBApproximationDistance(col,atari_palette[i]);
//		distance=RGBEuclidianDistance(col,atari_palette[i]);
		if (distance<min_distance)
		{
			min_distance=distance;
			most_similar=i;
		}
	}
	return most_similar;
}

int store_count=0;
unsigned char to_store[8];

void StoreGR11Pixel(unsigned char val)
{
	unsigned char pix=0;
	to_store[store_count]=val;
	store_count++;
	if (store_count==2)
	{
		pix |= to_store[0]<<4;
		pix |= to_store[1];
		fwrite(&pix,1,1,fp);
		store_count=0;
	}
}

void StoreGR15Pixel(unsigned char val)
{
	unsigned char pix=0;
	to_store[store_count]=val;
	store_count++;
	if (store_count==4)
	{
		pix |= to_store[0]<<6;
		pix |= to_store[1]<<4;
		pix |= to_store[2]<<2;
		pix |= to_store[3];
		fwrite(&pix,1,1,fp);
		store_count=0;
	}
}

void StoreGR8Pixel(unsigned char val)
{
	unsigned char pix=0;
	to_store[store_count]=val;
	store_count++;
	if (store_count==8)
	{
		pix |= to_store[0]<<7;
		pix |= to_store[1]<<6;
		pix |= to_store[2]<<5;
		pix |= to_store[3]<<4;
		pix |= to_store[4]<<3;
		pix |= to_store[5]<<2;
		pix |= to_store[6]<<1;
		pix |= to_store[7];
		fwrite(&pix,1,1,fp);
		store_count=0;
	}
}

bool exact=false;
int mode_1=15;
int mode_2=11;
int param_1=-1;
int param_2=-1;
int screen_scale=2;
int width, height;
int threshold;
int COLOR1;
int COLOR2;
int window_x;
int window_y;

string input_name;

char *output_name="out.bmp";
char *palette_name="palette.act";

ULONG_PTR GetParentProcessId() // By Napalm @ NetCore2K
{
 ULONG_PTR pbi[6];
 ULONG ulSize = 0;
 LONG (WINAPI *NtQueryInformationProcess)(HANDLE ProcessHandle, ULONG ProcessInformationClass,
  PVOID ProcessInformation, ULONG ProcessInformationLength, PULONG ReturnLength); 
 *(FARPROC *)&NtQueryInformationProcess = 
  GetProcAddress(LoadLibraryA("NTDLL.DLL"), "NtQueryInformationProcess");
 if(NtQueryInformationProcess){
  if(NtQueryInformationProcess(GetCurrentProcess(), 0,
    &pbi, sizeof(pbi), &ulSize) >= 0 && ulSize == sizeof(pbi))
     return pbi[5];
 }
 return (ULONG_PTR)-1;
}

void parse_params(int argc, char *argv[])
{
	parser.parse(argc,argv);

	if (argc<2)
	{
		if (!AttachConsole(GetParentProcessId()))
			AllocConsole();

		int hConHandle;
		long lStdHandle;
		FILE *fp_out;
		lStdHandle = (long)GetStdHandle(STD_OUTPUT_HANDLE);
		hConHandle = _open_osfhandle(lStdHandle, _O_TEXT);
		fp_out = _fdopen( hConHandle, "w" );
		*stdout = *fp_out;
		setvbuf( stdout, NULL, _IONBF, 0 );

		printf("\n\nUsage:\n");
		printf("PicMixer file.bmp [/zoom=val] [/exact] [/mode1=val] [/mode2=val] [/param1=val] [/param1=val]\n\n");
		exit(0);
	}

	string p_mode1 = parser.getValue("mode1","15");
	mode_1=String2Value<int>(p_mode1);

	string p_mode2 = parser.getValue("mode2","9");
	mode_2=String2Value<int>(p_mode2);

	string p_param1 = parser.getValue("param1","-1");
	param_1=String2Value<int>(p_param1);

	string p_param2 = parser.getValue("param2","-1");
	param_2=String2Value<int>(p_param2);

	string p_zoom = parser.getValue("zoom","2");
	screen_scale=String2Value<int>(p_zoom);

	string p_width = parser.getValue("width","-1");
	width=String2Value<int>(p_width);

	string p_height = parser.getValue("height","-1");
	height=String2Value<int>(p_height);

	string p_threshold = parser.getValue("threshold","7");
	threshold=String2Value<int>(p_threshold);

	string p_color1 = parser.getValue("color1","15");
	COLOR1=String2Value<int>(p_color1);

	string p_color2 = parser.getValue("color2","0");
	COLOR2=String2Value<int>(p_color2);

	string p_window_x = parser.getValue("window_x","0");
	window_x=String2Value<int>(p_window_x);

	string p_window_y = parser.getValue("window_y","0");
	window_y=String2Value<int>(p_window_y);

	exact = parser.switchExists("exact");

	input_name = string(argv[1]);
}

int convert()
{
	int x,y,pixel,atari_color,index,val;
	rgb col;

	int current_mode;
	int previous_mode;
	int current_param;
	int pixel_size;

	fp=fopen("pic.bin","wb+");
	if (!fp)
		return 1;

	if ( (mode_1==9 && mode_2==11) || 	// APAC simulation
		 (mode_1==11 && mode_2==9)
		 )
	{
		for (y=0;y<height;y+=2)
		{
			for (x=0;x<width;)
			{
				col.r=0;
				col.g=0;
				col.b=0;

				for (index=0;index<4;++index)
				{
					pixel=getpixel(bmp1,x+index,y);
					col.r += getr(pixel);
					col.g += getg(pixel);
					col.b += getb(pixel);

					pixel=getpixel(bmp1,x+index,y+1);
					col.r += getr(pixel);
					col.g += getg(pixel);
					col.b += getb(pixel);
				}
				col.r/=8;
				col.g/=8;
				col.b/=8;

				atari_color=FindAtariColorIndex(col);

//				StoreGR11Pixel(i/16); - to be done properly

				atari_color/=2; // use only 128 colors
				atari_color*=2;
				col=atari_palette[atari_color];
				for (index=0;index<4;++index,++x)
				{
					putpixel(out,x,y,makecol(col.r,col.g,col.b));
					putpixel(out,x,y+1,makecol(0,0,0));
				}
			}
		}
	}
	else
	for (y=0;y<height;y++)
	{
		for (x=0;x<width;)
		{
			col.r=0;
			col.g=0;
			col.b=0;

			if (y%2==0)
			{
				current_mode=mode_1;
				previous_mode=mode_2;
				current_param=param_1;
			}
			else
			{
				current_mode=mode_2;
				previous_mode=mode_1;
				current_param=param_2;
			}

			if (current_param==-1)
			{
				if (current_mode==9)
					current_param=0;
				else if (current_mode==10)
					current_param=7;
				else if (current_mode==11)
					current_param=8;
				else if (current_mode==15)
					current_param=0;
				else if (current_mode==8)
					current_param=15;
			}

			switch(current_mode)
			{
				case 11:
				case 10:
				case 9:
					pixel_size=4;
					break;
				case 15:
					pixel_size=2;
					break;
				case 8:
					pixel_size=1;
					break;
			}

			if (current_mode==previous_mode || exact)
			{
				for (index=0;index<pixel_size;++index)
				{
					pixel=getpixel(bmp1,x+index,y);
					col.r += getr(pixel);
					col.g += getg(pixel);
					col.b += getb(pixel);

				}
				col.r/=pixel_size;
				col.g/=pixel_size;
				col.b/=pixel_size;
			}
			else
			{
				for (index=0;index<pixel_size;++index)
				{
					pixel=getpixel(bmp1,x+index,y);
					col.r += getr(pixel);
					col.g += getg(pixel);
					col.b += getb(pixel);

					if (previous_mode==10)
						pixel=getpixel(bmp1,x+index,y+1*(y%2?-1:1));
					else
						pixel=getpixel(bmp1,x+index,y+1*(y%2?-1:1));
					col.r += getr(pixel);
					col.g += getg(pixel);
					col.b += getb(pixel);
				}
				col.r/=pixel_size*2;
				col.g/=pixel_size*2;
				col.b/=pixel_size*2;
			}

			atari_color=FindAtariColorIndex(col);

			if (current_mode==11)
			{
				// gr.11
				if (atari_color>4 && atari_color<16) // we don't have grays in gr. 11, so change gray to other, closest
				{
/*
					float h,s,v;
					rgb_to_hsv(col.r,col.g,col.b,&h,&s,&v);
					s=1;
					v=0.5;

					hsv_to_rgb(h,s,v,&col.r,&col.g,&col.b);
					i=FindAtariColorIndex(col);
*/
					if (atari_color>4 && atari_color<16) // we don't have grays in gr. 11, so change gray to other
						atari_color+=16*current_param;
				}

				atari_color-=atari_color%16; // gr. 11

				StoreGR11Pixel(atari_color/16);

				if (atari_color!=0)// gr. 11
					atari_color+=7;// gr. 11

				col=atari_palette[atari_color];
				for (index=0;index<4;++index,++x)
					putpixel(out,x,y,makecol(col.r,col.g,col.b));
			}
			else if (current_mode==9)
			{
				atari_color=atari_color%16;  // gr.9

				StoreGR11Pixel(atari_color);

				atari_color+=16*current_param; // gr.9

				col=atari_palette[atari_color];
				for (index=0;index<4;++index,++x)
					putpixel(out,x,y,makecol(col.r,col.g,col.b));
			}
			else if (current_mode==10) // 8 (9) greys
			{
				atari_color=atari_color%16;  // gr.10
				atari_color/=2;
				atari_color*=2;

				StoreGR15Pixel(atari_color/2);

				atari_color+=16*current_param; // gr.10

				col=atari_palette[atari_color];
				for (index=0;index<4;++index,++x)
					putpixel(out,x+2,y,makecol(col.r,col.g,col.b));
			}
			else if (current_mode==15)
			{
				atari_color=atari_color%16;  // gr.15
				if (atari_color<3)
				{
					atari_color=0;
					val=0;
				}
				else if (atari_color<8)
				{
					atari_color=5;
					val=1;
				}
				else if (atari_color<13)
				{
					atari_color=10;
					val=2;
				}
				else 
				{
					atari_color=15;
					val=3;
				}

				StoreGR15Pixel(val);

				col=atari_palette[atari_color+current_param*16];
				for (index=0;index<2;++index,++x)
					putpixel(out,x,y,makecol(col.r,col.g,col.b));		
			}
			else if (current_mode==8)
			{
				// convert to brightness
				atari_color=atari_color%16;  // gr.8
				if (atari_color<threshold)
				{
					val=0;
					atari_color=COLOR2;
				}
				else
				{
					val=1;
					atari_color=COLOR1;
				}
				StoreGR8Pixel(val);
				col=atari_palette[atari_color];
				putpixel(out,x++,y,makecol(col.r,col.g,col.b));		
			}
			else 	
			{
				printf("Unsupported mode!\n");
				exit(0);
			}
		} // for x
	} // for y

	save_bitmap(output_name,out,NULL);
	fclose(fp);
	return 0;
}

int convert_window_to_bitmap()
{
	bmp1=create_bitmap(width,height);
	out=create_bitmap(width,height);

	HDC hdc;
	HWND window_handle;

	for (;;)
	{
		// 1. Choose Window 
		for (;;)
		{
			POINT point;
			GetCursorPos(&point);
			window_handle = WindowFromPoint(point);
			hdc = GetWindowDC(window_handle);

			blit_from_hdc(hdc,bmp1,window_x,window_y,0,0,width,height);
			textout_ex(bmp1,font,"Hold SHIFT to lock on this window",0,0,makecol(255, 255, 255),0);
			textout_ex(bmp1,font,"Press ESCAPE to unlock",0,8,makecol(255, 255, 255),0);
			textout_ex(bmp1,font,"Press Q to exit",0,16,makecol(255, 255, 255),0);
			stretch_blit(bmp1,screen,0,0,width,height,0,0,screen->w,screen->h);

			if (GetAsyncKeyState(VK_LSHIFT)&1)
				break;
			if (key[KEY_Q])
				return 0;
			Sleep(5);
		}
		for (;;)
		{
			if (IsWindow(window_handle))
			{
				blit_from_hdc(hdc,bmp1,window_x,window_y,0,0,width,height);
				
				convert();
				stretch_blit(out,screen,0,0,width,height,0,0,screen->w,screen->h);
			}
			else
			{
				break;
			}
			Sleep(10);
			if (key[KEY_ESC])
				break;
			if (key[KEY_Q])
				return 0;
		}
	}
	return 0;
}


int convert_from_file()
{
	bmp1=load_bitmap(input_name.c_str(),NULL);
	if (bmp1==NULL)
	{
		printf("Error loading input bitmap!\n");
		exit(0);
	}
	if (width==-1)
		width=bmp1->w;
	if (height==-1)
		height=bmp1->h;

	out=create_bitmap(width,height);

	convert();
	return 0;
}

int main(int argc, char *argv[])
{
	parse_params(argc,argv);
	if (input_name=="window")
	{
		if (width==-1)
			width=320;
		if (height==-1)
			height=240;
	}
	allegro_init(); // Initialize Allegro
	install_keyboard(); 
	set_color_depth(desktop_color_depth());
	set_gfx_mode(GFX_AUTODETECT_WINDOWED, width*screen_scale,height*screen_scale,0,0); // Change our graphics mode to 640x480
	set_display_switch_mode(SWITCH_BACKGROUND);
	

	LoadAtariPalette(palette_name);

	if (input_name=="window")
		convert_window_to_bitmap();
	else
		convert_from_file();
	return 0;
}

END_OF_MAIN();
